use std::{fs::File, path::Path};

use eyre::Context;
use rayon::prelude::{IndexedParallelIterator, ParallelIterator};

use crate::{
    chunk::{read_chunks_data, read_chunks_messages, MessageData},
    index::BagIndex,
    info::BagInfo,
    message::MessageLayouts,
    reader::MmapReader,
    Result,
};

pub struct Bag {
    reader: MmapReader,
    index: BagIndex,
}

impl Bag {
    pub fn open<P: AsRef<Path>>(path: P) -> Result<Self> {
        let bag_file = File::open(path).wrap_err("could not open bag file")?;
        let mut reader = MmapReader::map(bag_file).wrap_err("could not map bag file memory")?;

        // TODO: support unindexed bags
        let index = BagIndex::read(&mut reader).wrap_err("error reading bag index")?;
        Ok(Bag { reader, index })
    }

    pub fn index(&self) -> &BagIndex {
        &self.index
    }

    pub fn reader(&self) -> &MmapReader {
        &self.reader
    }

    pub fn compute_message_layouts(&mut self) -> Result<MessageLayouts> {
        MessageLayouts::new(&self.index.connections)
    }

    pub fn compute_info(&mut self) -> Result<BagInfo> {
        BagInfo::compute(self.reader.clone(), &self.index)
    }

    pub fn read_chunks(&mut self) -> impl IndexedParallelIterator<Item = Result<Vec<u8>>> + '_ {
        read_chunks_data(self.reader.clone(), &self.index.chunks)
    }

    pub fn read_messages(&mut self) -> impl ParallelIterator<Item = Result<MessageData>> + '_ {
        read_chunks_messages(self.reader.clone(), &self.index.chunks)
    }
}
