use std::{
    io::{self, Read, SeekFrom},
    mem,
    str::FromStr,
};

use bytes::Bytes;
use eyre::{bail, eyre, Context};
use rayon::{
    iter::{IntoParallelIterator, ParallelIterator},
    prelude::IndexedParallelIterator,
};

use crate::{
    error,
    index::ChunkInfo,
    parse::{Header, Op},
    reader::{BagReader, BytesReader},
    Error, Result, Time,
};

#[derive(Clone, Copy, Debug)]
pub enum Compression {
    None,
    Bz2,
    Lz4,
}

impl FromStr for Compression {
    type Err = Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(match s {
            "none" => Compression::None,
            "bz2" => Compression::Bz2,
            "lz4" => Compression::Lz4,
            _ => bail!(error::UnsupportedCompression(s.to_string())),
        })
    }
}

impl Compression {
    pub fn decompress<R: io::Read>(self, mut read: R, decompressed_size: usize) -> Result<Vec<u8>> {
        let mut decompressed = Vec::with_capacity(decompressed_size);
        match self {
            Compression::None => {
                read.read_to_end(&mut decompressed)?;
            }
            Compression::Bz2 => todo!("bz2 decompression"),
            Compression::Lz4 => {
                lz4_flex::frame::FrameDecoder::new(read).read_to_end(&mut decompressed)?;
            }
        }
        Ok(decompressed)
    }
}

#[derive(Debug)]
pub struct ChunkHeader {
    pub compression: Compression,
    pub uncompressed_size: u32,
}

impl ChunkHeader {
    pub fn read<R: BagReader>(reader: &mut R) -> Result<Self> {
        let header = reader.read_header_op(Op::Chunk)?;
        Self::from_header(header)
    }

    pub fn from_header(header: Header) -> Result<Self> {
        Ok(Self {
            compression: header.read_string(b"compression")?.parse()?,
            uncompressed_size: header.read_u32(b"size")?,
        })
    }
}

pub fn read_chunk_data<R: BagReader>(bag_reader: &mut R) -> Result<Vec<u8>> {
    let chunk_header = ChunkHeader::read(bag_reader)?;

    let compressed_data = bag_reader.read_data()?;
    let data = chunk_header
        .compression
        .decompress(compressed_data, chunk_header.uncompressed_size as usize)?;

    Ok(data)
}

pub fn read_chunk_data_at<R: BagReader + io::Seek>(
    bag_reader: &mut R,
    pos: u64,
) -> Result<Vec<u8>> {
    bag_reader.seek(SeekFrom::Start(pos as u64))?;
    read_chunk_data(bag_reader).wrap_err_with(|| eyre!("failed to read chunk at offset {}", pos))
}

pub fn read_chunks_data<'a, R, C>(
    reader: R,
    chunks: C,
) -> impl IndexedParallelIterator<Item = Result<Vec<u8>>> + 'a
where
    R: BagReader + io::Seek + Clone + Send + 'a,
    C: IntoParallelIterator<Item = &'a ChunkInfo> + 'a,
    <C as IntoParallelIterator>::Iter: IndexedParallelIterator,
{
    chunks
        .into_par_iter()
        .map_with(reader, move |reader, chunk| {
            read_chunk_data_at(reader, chunk.pos)
        })
}

#[derive(Debug)]
pub struct MessageDataHeader {
    pub conn_id: u32,
    pub time: Time,
}

impl MessageDataHeader {
    pub fn read<R: BagReader>(reader: &mut R) -> Result<Self> {
        let header = reader.read_header_op(Op::MsgData)?;
        Self::from_header(header)
    }

    pub fn from_header(header: Header) -> Result<Self> {
        Ok(Self {
            conn_id: header.read_u32(b"conn")?,
            time: header.read_time(b"time")?,
        })
    }
}

#[derive(Debug)]
pub struct MessageData {
    pub header: MessageDataHeader,
    pub data: Bytes,
}

pub fn read_chunks_messages<'a, R, C>(
    reader: R,
    chunks: C,
) -> impl ParallelIterator<Item = Result<MessageData>> + 'a
where
    R: BagReader + io::Seek + Clone + Send + 'a,
    C: IntoParallelIterator<Item = &'a ChunkInfo> + 'a,
    <C as IntoParallelIterator>::Iter: IndexedParallelIterator,
{
    read_chunks_data(reader, chunks).flat_map_iter(move |data| ChunkMessageIterator {
        reader: data.map(|data| BytesReader::from(Bytes::from(data))),
    })
}

pub struct ChunkMessageIterator {
    reader: Result<BytesReader>,
}

impl ChunkMessageIterator {
    pub fn new<B: Into<Bytes>>(b: B) -> Self {
        Self {
            reader: Ok(BytesReader::from(b.into())),
        }
    }

    fn next_impl(&mut self) -> Result<Option<MessageData>> {
        let reader = match &mut self.reader {
            Ok(reader) => reader,
            Err(err) => {
                // workaround for eyre::Report not being clone
                let mut new_err = eyre!("original error already consumed");
                mem::swap(err, &mut new_err);
                return Err(new_err);
            }
        };
        while reader.remaining() > 0 {
            let header = reader.read_header()?;
            let op = header.read_op()?;

            match op {
                Op::MsgData => {
                    let header = MessageDataHeader::from_header(header)?;
                    let data = reader.read_data_bytes()?;
                    return Ok(Some(MessageData { header, data }));
                }
                Op::Connection => {
                    reader.skip_data()?;
                }
                _ => bail!("unexpected op in chunk: {:?}", op),
            }
        }
        Ok(None)
    }
}

impl Iterator for ChunkMessageIterator {
    type Item = Result<MessageData>;

    fn next(&mut self) -> Option<Self::Item> {
        self.next_impl().transpose()
    }
}
