use std::io;

use eyre::bail;
use log::trace;
use nom::{number::streaming::le_u32, sequence::tuple, Parser};

use crate::{
    error,
    parse::{self, header::fields::parse_time, Header, Op},
    reader::BagReader,
    Result, Time,
};

#[derive(Debug, Clone)]
pub struct ConnInfo {
    pub id: u32,
    pub topic: String,
    pub message_definition: String,
    pub datatype: String,
    pub md5sum: String,
}

impl ConnInfo {
    pub fn read<R: BagReader>(reader: &mut R) -> Result<Self> {
        let record_header = reader.read_header_op(Op::Connection)?;
        let conn_header = reader.read_header()?;
        Self::from_headers(record_header, conn_header)
    }

    pub fn from_headers(record_header: Header, conn_header: Header) -> Result<Self> {
        Ok(ConnInfo {
            id: record_header.read_u32(b"conn")?,
            topic: record_header.read_string(b"topic")?,
            message_definition: conn_header.read_string(b"message_definition")?,
            datatype: conn_header.read_string(b"type")?,
            md5sum: conn_header.read_string(b"md5sum")?,
        })
    }
}

#[derive(Clone, Debug)]
pub struct ChunkInfo {
    pub pos: u64,
    pub start_time: Time,
    pub end_time: Time,
    pub connections: Vec<ChunkConnection>,
}

impl ChunkInfo {
    pub fn read<R: BagReader>(reader: &mut R) -> Result<Self> {
        let header = reader.read_header_op(Op::ChunkInfo)?;
        if header.read_u32(b"ver")? != 1 {
            bail!("unsupported ChunkInfo version");
        }

        reader.read_data_length()?; // Data length not needed

        let conn_count = header.read_u32(b"count")?;
        let mut connections = Vec::with_capacity(conn_count as usize);
        for _ in 0..conn_count {
            let conn = ChunkConnection::read(reader)?;
            connections.push(conn);
        }
        Ok(ChunkInfo {
            pos: header.read_u64(b"chunk_pos")?,
            start_time: header.read_time(b"start_time")?,
            end_time: header.read_time(b"end_time")?,
            connections,
        })
    }
}

#[derive(Clone, Debug)]
pub struct ChunkConnection {
    pub conn_id: u32,
    pub count: u32,
}

impl ChunkConnection {
    pub fn parse(input: parse::Input) -> parse::IResult<Self> {
        tuple((le_u32, le_u32))
            .map(|(conn_id, count)| ChunkConnection { conn_id, count })
            .parse(input)
    }

    pub fn read<R: BagReader>(reader: &mut R) -> Result<Self> {
        reader.read_parser(Self::parse)
    }
}

#[derive(Clone, Debug)]
pub struct BagIndex {
    pub connections: Vec<ConnInfo>,
    pub chunks: Vec<ChunkInfo>,
}

impl BagIndex {
    fn read_v2<R: BagReader + io::Seek>(reader: &mut R) -> Result<Self> {
        let file_header = reader.read_header_op(Op::FileHeader)?;

        let data_length = reader.read_data_length()?;
        trace!("data length: {}", data_length);

        let index_pos = file_header.read_u64(b"index_pos")?;
        trace!("index pos: {}", index_pos);

        if index_pos == 0 {
            bail!(error::Unindexed);
        }

        if let Ok(encryptor) = file_header.read_string(b"encryptor") {
            bail!(error::UnsupportedEncryptor(encryptor));
        }

        reader.seek(io::SeekFrom::Start(index_pos))?;

        let conn_count = file_header.read_u32(b"conn_count")?;
        trace!("connection count: {}", conn_count);

        let mut connections = Vec::with_capacity(conn_count as usize);

        for _ in 0..conn_count {
            let conn = ConnInfo::read(reader)?;
            trace!("connection: id={}, topic={}", conn.id, conn.topic);
            connections.push(conn);
        }

        let chunk_count = file_header.read_u32(b"chunk_count")?;
        trace!("chunk count: {}", chunk_count);

        let mut chunks = Vec::with_capacity(chunk_count as usize);

        for _ in 0..chunk_count {
            let chunk = ChunkInfo::read(reader)?;
            chunks.push(chunk);
        }

        Ok(BagIndex {
            connections,
            chunks,
        })
    }

    pub fn read<R: BagReader + io::Seek>(reader: &mut R) -> Result<Self> {
        let version = reader.read_version()?;
        trace!("bag version: {}", version);
        if (version.major, version.minor) == (2, 0) {
            Self::read_v2(reader)
        } else {
            bail!(error::UnsupportedVersion(version));
        }
    }
}

#[derive(Debug)]
pub struct IndexData {
    pub conn_id: u32,
    pub entries: Vec<IndexEntry>,
}

impl IndexData {
    pub fn read<R: BagReader>(reader: &mut R) -> Result<Self> {
        let header = reader.read_header_op(Op::IndexData)?;
        if header.read_u32(b"ver")? != 1 {
            bail!("unsupported IndexData version");
        }

        reader.read_data_length()?; // Data length not needed

        let entry_count = header.read_u32(b"count")?;
        let mut entries = Vec::with_capacity(entry_count as usize);
        for _ in 0..entry_count {
            let conn = IndexEntry::read(reader)?;
            entries.push(conn);
        }
        Ok(IndexData {
            conn_id: header.read_u32(b"conn")?,
            entries,
        })
    }
}

#[derive(Debug)]
pub struct IndexEntry {
    pub time: Time,
    pub offset: u32,
}

impl IndexEntry {
    pub fn parse(input: parse::Input) -> parse::IResult<Self> {
        tuple((parse_time, le_u32))
            .map(|(time, offset)| IndexEntry { time, offset })
            .parse(input)
    }

    pub fn read<R: BagReader>(reader: &mut R) -> Result<Self> {
        reader.read_parser(Self::parse)
    }
}
