use std::collections::HashMap;

use eyre::{bail, eyre};
use ros_message::{DataType, FieldCase, FieldInfo, I8Variant, MessagePath, Msg, U8Variant};
use thiserror::Error;

use crate::{Result, SmallStr};

type Dependencies = HashMap<MessagePath, MessageLayoutRef>;

#[derive(Clone, Debug, Error)]
#[error("missing dependent message definition: {0}")]
pub struct MissingDependency(MessagePath);

#[derive(Clone, Debug, Error)]
#[error("dependency cycle involving: {0:?}")]
pub struct DependencyCycle(Vec<String>);

struct Resolver<'a> {
    dependencies: &'a HashMap<MessagePath, MessageLayoutRef>,
    current_path: &'a MessagePath,
}

impl<'a> Resolver<'a> {
    fn get_dep(&self, path: &MessagePath) -> Result<MessageLayoutRef> {
        self.dependencies
            .get(path)
            .cloned()
            .ok_or_else(|| eyre!(MissingDependency(path.clone())))
    }

    fn get_local_dep(&self, name: &str) -> Result<MessageLayoutRef> {
        let full_path = self.current_path.peer(name);
        self.get_dep(&full_path)
    }
}

#[derive(Clone, Debug)]
pub enum FieldType {
    /// Represents `bool`.
    Bool,
    /// Represents `int8` or `byte`.
    I8(I8Variant),
    /// Represents `int16`.
    I16,
    /// Represents `int32`.
    I32,
    /// Represents `int64`.
    I64,
    /// Represents `uint8` or `char`.
    U8(U8Variant),
    /// Represents `uint16`.
    U16,
    /// Represents `uint32`.
    U32,
    /// Represents `uint64`.
    U64,
    /// Represents `float32`.
    F32,
    /// Represents `float64`.
    F64,
    /// Represents `string`.
    String,
    /// Represents `time`.
    Time,
    /// Represents `duration`.
    Duration,
    // Another messgage type.
    Message(MessageLayoutRef),
}

impl FieldType {
    fn from_datatype_and_resolver(datatype: &DataType, resolver: &Resolver) -> Result<Self> {
        Ok(match datatype {
            DataType::Bool => FieldType::Bool,
            DataType::I8(v) => FieldType::I8(v.clone()),
            DataType::I16 => FieldType::I16,
            DataType::I32 => FieldType::I32,
            DataType::I64 => FieldType::I64,
            DataType::U8(v) => FieldType::U8(v.clone()),
            DataType::U16 => FieldType::U16,
            DataType::U32 => FieldType::U32,
            DataType::U64 => FieldType::U64,
            DataType::F32 => FieldType::F32,
            DataType::F64 => FieldType::F64,
            DataType::String => FieldType::String,
            DataType::Time => FieldType::Time,
            DataType::Duration => FieldType::Duration,
            DataType::LocalMessage(path) => FieldType::Message(resolver.get_local_dep(path)?),
            DataType::GlobalMessage(path) => FieldType::Message(resolver.get_dep(path)?),
        })
    }
}

#[derive(Clone, Debug)]
pub struct FieldLayout {
    pub typ: FieldType,
    pub name: SmallStr,
    pub multiplicity: Multiplicity,
}

impl FieldLayout {
    fn from_info_and_resolver(info: &FieldInfo, resolver: &Resolver) -> Result<Self> {
        Ok(Self {
            typ: FieldType::from_datatype_and_resolver(info.datatype(), resolver)?,
            name: info.name().into(),
            multiplicity: match info.case() {
                FieldCase::Unit => Multiplicity::Unit,
                FieldCase::Vector => Multiplicity::Dynamic,
                FieldCase::Array(n) => Multiplicity::Fixed(*n),
                FieldCase::Const(_) => unreachable!(),
            },
        })
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub enum Multiplicity {
    Unit,
    Fixed(usize),
    Dynamic,
}

#[derive(Clone, Debug)]
pub struct MessageLayout {
    fields: Vec<FieldLayout>,
}

pub type MessageLayoutRef = std::sync::Arc<MessageLayout>;

impl MessageLayout {
    pub fn fields(&self) -> &[FieldLayout] {
        &self.fields
    }

    pub fn from_msgs<'a, M, I>(msgs: M) -> Result<MessageLayout>
    where
        M: IntoIterator<IntoIter = I>,
        I: Iterator<Item = &'a Msg> + DoubleEndedIterator,
    {
        let mut dependencies: Dependencies = Default::default();
        let mut msgs = msgs.into_iter();
        let main = msgs
            .next()
            .ok_or_else(|| eyre!("at least one msg required"))?;

        let mut deps: Vec<_> = msgs.collect();

        while !deps.is_empty() {
            let mut remaining = Vec::new();
            let orig_length = deps.len();
            for msg in deps.drain(..) {
                let layout = match MessageLayout::from_msg_and_deps(msg, &dependencies) {
                    Ok(layout) => layout,
                    Err(err) => {
                        if err.root_cause().is::<MissingDependency>() {
                            // Has a missing dependency, resovle it on the next iteration
                            remaining.push(msg);
                            continue;
                        } else {
                            return Err(err);
                        }
                    }
                };
                dependencies.insert(msg.path().clone(), MessageLayoutRef::from(layout));
            }
            if remaining.len() == orig_length {
                // Made no progress, there is a dependency cycle
                bail!(DependencyCycle(
                    remaining.iter().map(|msg| msg.path().to_string()).collect()
                ));
            }
            deps = remaining;
        }

        MessageLayout::from_msg_and_deps(main, &dependencies)
    }

    fn from_msg_and_deps(msg: &Msg, dependencies: &Dependencies) -> Result<MessageLayout> {
        let resolver = Resolver {
            dependencies,
            current_path: msg.path(),
        };
        let fields = msg
            .fields()
            .iter()
            .filter(|field| !field.is_constant())
            .map(|field| FieldLayout::from_info_and_resolver(field, &resolver))
            .collect::<Result<Vec<_>, _>>()?;
        Ok(MessageLayout { fields })
    }
}
