use std::{collections::HashMap, convert::TryFrom, sync::Arc};

use eyre::eyre;
use lazy_static::lazy_static;
use log::trace;
use regex::Regex;
use ros_message::{MessagePath, MessageValue, Msg};

use crate::{
    chunk::MessageData, index::ConnInfo, layout::MessageLayout, parse::parse_message, Result,
};

fn parse_msg(message_name: &str, msgdef: &str) -> Result<Msg> {
    trace!("message definition {}: {}", message_name, msgdef);
    let path = MessagePath::try_from(message_name)?;
    let msgtype = Msg::new(path, msgdef)?;
    Ok(msgtype)
}

lazy_static! {
    static ref BOUNDARY_RE: Regex = Regex::new(r"\r?\n==+\r?\nMSG: ([^\r\n]+)\r?\n").unwrap();
}

fn parse_msgs(conn: &ConnInfo) -> Result<Vec<Msg>> {
    let msgdefs = &conn.message_definition;
    let mut name = conn.datatype.clone();
    let mut begin = 0usize;
    let mut msgs = Vec::new();

    for cap in BOUNDARY_RE.captures_iter(msgdefs) {
        let boundary_range = cap.get(0).unwrap();
        let end = boundary_range.start();

        let msgdef = &msgdefs[begin..end];
        let msgtype = parse_msg(&name, msgdef)?;
        msgs.push(msgtype);

        name = cap[1].to_string();
        begin = boundary_range.end();
    }

    let msgdef = &msgdefs[begin..];

    let msg = parse_msg(&name, msgdef)?;
    msgs.push(msg);

    Ok(msgs)
}

pub fn compute_layout(conn: &ConnInfo) -> Result<MessageLayout> {
    let msgs = parse_msgs(conn)?;
    MessageLayout::from_msgs(&msgs)
}

pub struct ConnData {
    pub info: Arc<ConnInfo>,
    pub layout: MessageLayout,
}

impl ConnData {
    fn new(conn: &ConnInfo) -> Result<ConnData> {
        let layout = compute_layout(conn)?;
        Ok(ConnData {
            info: Arc::new(conn.clone()),
            layout,
        })
    }
}
pub struct MessageLayouts {
    pub conns: HashMap<u32, ConnData>,
}

impl MessageLayouts {
    pub fn new<'a, I: IntoIterator<Item = &'a ConnInfo>>(connections: I) -> Result<MessageLayouts> {
        let mut conns = HashMap::new();
        for conn in connections {
            conns.insert(conn.id, ConnData::new(conn)?);
        }
        Ok(MessageLayouts { conns })
    }

    pub fn get(&self, conn_id: u32) -> Option<&MessageLayout> {
        self.conns.get(&conn_id).map(|conn_data| &conn_data.layout)
    }

    pub fn decode(&self, message: &MessageData) -> Result<(Arc<ConnInfo>, MessageValue)> {
        let conn = self
            .conns
            .get(&message.header.conn_id)
            .ok_or_else(|| eyre!("missing connection: {}", message.header.conn_id))?;

        let value = parse_message(&conn.layout, message.data.as_ref())?;

        Ok((conn.info.clone(), value))
    }

    #[cfg(feature = "bson")]
    pub fn decode_bson(&self, message: &MessageData) -> Result<(Arc<ConnInfo>, bson::Document)> {
        use crate::parse::parse_message_bson;

        let conn = self
            .conns
            .get(&message.header.conn_id)
            .ok_or_else(|| eyre!("missing connection: {}", message.header.conn_id))?;

        let value = parse_message_bson(&conn.layout, message.data.as_ref())?;

        Ok((conn.info.clone(), value))
    }
}
