use eyre::bail;
use nom::{
    combinator::map_res,
    multi::{length_count, length_data, many_m_n},
    number::complete::{
        le_f32, le_f64, le_i16, le_i32, le_i64, le_i8, le_u16, le_u32, le_u64, le_u8,
    },
    Parser,
};
use ros_message::{MessageValue, Value};

use crate::{
    layout::{FieldLayout, FieldType, MessageLayout, Multiplicity},
    parse::{header::fields::parse_time, ErrorKind},
};

use super::{Error, IResult, Input};
use crate::Result;

fn type_parser<'a, 'b: 'a>(
    typ: &'a FieldType,
) -> impl Parser<Input<'b>, Value, Error<Input<'b>>> + 'a {
    move |input| -> IResult<Value> {
        match typ {
            FieldType::Bool => le_u8.map(|v| Value::Bool(v != 0)).parse(input),
            FieldType::I8(_) => le_i8.map(Value::I8).parse(input),
            FieldType::I16 => le_i16.map(Value::I16).parse(input),
            FieldType::I32 => le_i32.map(Value::I32).parse(input),
            FieldType::I64 => le_i64.map(Value::I64).parse(input),
            FieldType::U8(_) => le_u8.map(Value::U8).parse(input),
            FieldType::U16 => le_u16.map(Value::U16).parse(input),
            FieldType::U32 => le_u32.map(Value::U32).parse(input),
            FieldType::U64 => le_u64.map(Value::U64).parse(input),
            FieldType::F32 => le_f32.map(Value::F32).parse(input),
            FieldType::F64 => le_f64.map(Value::F64).parse(input),
            FieldType::String => map_res(length_data(le_u32), |s| {
                String::from_utf8(Vec::from(s))
                    .map_err(ErrorKind::from)
                    .map(Value::String)
            })
            .parse(input),
            FieldType::Time => parse_time.map(Value::Time).parse(input),
            FieldType::Duration => todo!(),
            FieldType::Message(layout) => message_parser(layout.as_ref())
                .map(Value::Message)
                .parse(input),
        }
    }
}

fn field_parser<'a, 'b: 'a>(
    field: &'a FieldLayout,
) -> impl Parser<Input<'b>, (String, Value), Error<Input<'b>>> + 'a {
    move |input| -> IResult<(String, Value)> {
        let mut parser = type_parser(&field.typ);
        let (rest, value) = match field.multiplicity {
            Multiplicity::Unit => parser.parse(input),
            Multiplicity::Fixed(n) => many_m_n(n, n, parser).map(Value::Array).parse(input),
            Multiplicity::Dynamic => length_count(le_u32, parser).map(Value::Array).parse(input),
        }?;
        Ok((rest, (String::from(field.name.as_str()), value)))
    }
}

fn message_parser<'a, 'b: 'a>(
    layout: &'a MessageLayout,
) -> impl Parser<Input<'b>, MessageValue, Error<Input<'b>>> + 'a {
    move |mut input| -> IResult<_> {
        let mut message = MessageValue::with_capacity(layout.fields().len());
        for field in layout.fields() {
            let (rest, (name, value)) = field_parser(field).parse(input)?;
            input = rest;
            message.insert(name, value);
        }
        Ok((input, message))
    }
}

pub fn parse_message(layout: &MessageLayout, input: &[u8]) -> Result<MessageValue> {
    match message_parser(layout).parse(input) {
        Ok((&[], message)) => Ok(message),
        Ok(_) => bail!("extra data after message"),
        Err(nom::Err::Incomplete(_)) => unreachable!(),
        Err(nom::Err::Error(e) | nom::Err::Failure(e)) => Err(e.into_owned().into()),
    }
}
