use bson::{Bson, Document};
use eyre::bail;
use nom::{
    bytes::complete::take,
    combinator::map_res,
    multi::{length_count, length_data, many_m_n},
    number::complete::{
        le_f32, le_f64, le_i16, le_i32, le_i64, le_i8, le_u16, le_u32, le_u64, le_u8,
    },
    Parser,
};

use crate::{
    layout::{FieldLayout, FieldType, MessageLayout, Multiplicity},
    parse::{header::fields::parse_time, ErrorKind},
};

use super::{Error, IResult, Input};
use crate::Result;

fn type_parser<'a, 'b: 'a>(
    typ: &'a FieldType,
) -> impl Parser<Input<'b>, Bson, Error<Input<'b>>> + 'a {
    move |input| -> IResult<Bson> {
        match typ {
            FieldType::Bool => le_u8.map(|v| Bson::Boolean(v != 0)).parse(input),
            FieldType::I8(_) => le_i8.map(|v| Bson::Int32(v as i32)).parse(input),
            FieldType::I16 => le_i16.map(|v| Bson::Int32(v as i32)).parse(input),
            FieldType::I32 => le_i32.map(Bson::Int32).parse(input),
            FieldType::I64 => le_i64.map(Bson::Int64).parse(input),
            FieldType::U8(_) => le_u8.map(|v| Bson::Int32(v as i32)).parse(input),
            FieldType::U16 => le_u16.map(|v| Bson::Int32(v as i32)).parse(input),
            FieldType::U32 => le_u32.map(|v| Bson::Int64(v as i64)).parse(input),
            FieldType::U64 => le_u64.map(|v| Bson::Int64(v as i64)).parse(input),
            FieldType::F32 => le_f32.map(|v| Bson::Double(v as f64)).parse(input),
            FieldType::F64 => le_f64.map(Bson::Double).parse(input),
            FieldType::String => map_res(length_data(le_u32), |s| {
                String::from_utf8(Vec::from(s))
                    .map_err(ErrorKind::from)
                    .map(Bson::String)
            })
            .parse(input),
            FieldType::Time => parse_time
                .map(|t| bson::bson!({"sec": (t.sec as i64), "nsec": (t.nsec as i64)}))
                .parse(input),
            FieldType::Duration => todo!(),
            FieldType::Message(layout) => message_parser(layout.as_ref())
                .map(Bson::Document)
                .parse(input),
        }
    }
}

fn field_parser<'a, 'b: 'a>(
    field: &'a FieldLayout,
) -> impl Parser<Input<'b>, (String, Bson), Error<Input<'b>>> + 'a {
    move |input| -> IResult<(String, Bson)> {
        let mut parser = type_parser(&field.typ);
        let (rest, value) = match field.multiplicity {
            Multiplicity::Unit => parser.parse(input),
            Multiplicity::Fixed(n) => match field.typ {
                FieldType::U8(_) => take(n)
                    .map(|b| bson::Binary {
                        subtype: bson::spec::BinarySubtype::Generic,
                        bytes: Vec::from(b),
                    })
                    .map(Bson::from)
                    .parse(input),
                _ => many_m_n(n, n, parser).map(Bson::Array).parse(input),
            },
            Multiplicity::Dynamic => match field.typ {
                FieldType::U8(_) => length_data(le_u32)
                    .map(|b| bson::Binary {
                        subtype: bson::spec::BinarySubtype::Generic,
                        bytes: Vec::from(b),
                    })
                    .map(Bson::from)
                    .parse(input),
                _ => length_count(le_u32, parser).map(Bson::Array).parse(input),
            },
        }?;
        Ok((rest, (String::from(field.name.as_str()), value)))
    }
}

fn message_parser<'a, 'b: 'a>(
    layout: &'a MessageLayout,
) -> impl Parser<Input<'b>, Document, Error<Input<'b>>> + 'a {
    move |mut input| -> IResult<_> {
        let mut message = Document::new();
        for field in layout.fields() {
            let (rest, (name, value)) = field_parser(field).parse(input)?;
            input = rest;
            message.insert(name, value);
        }
        Ok((input, message))
    }
}

pub fn parse_message_bson(layout: &MessageLayout, input: &[u8]) -> Result<Document> {
    match message_parser(layout).parse(input) {
        Ok((&[], message)) => Ok(message),
        Ok(_) => bail!("extra data after message"),
        Err(nom::Err::Incomplete(_)) => unreachable!(),
        Err(nom::Err::Error(e) | nom::Err::Failure(e)) => Err(e.into_owned().into()),
    }
}
