use bytes::{Buf, Bytes};
use eyre::bail;

use super::{error::UnexpectedEof, BagReader};
use crate::{parse, Result};

// This is like a SliceReader, except a byte is reference
// counted to read_data_bytes can give out unlimited-lifetime references
// to the same buffer.
#[derive(Clone)]
pub struct BytesReader {
    current: Bytes,
    last: Option<Bytes>,
}

impl BytesReader {
    pub fn into_inner(self) -> Bytes {
        self.current
    }

    pub fn remaining(&self) -> usize {
        self.as_ref().remaining()
    }
}

impl AsRef<Bytes> for BytesReader {
    fn as_ref(&self) -> &Bytes {
        &self.current
    }
}

impl AsMut<Bytes> for BytesReader {
    fn as_mut(&mut self) -> &mut Bytes {
        &mut self.current
    }
}

impl From<Bytes> for BytesReader {
    fn from(bytes: Bytes) -> Self {
        Self {
            current: bytes,
            last: None,
        }
    }
}

impl BagReader for BytesReader {
    fn skip(&mut self, amount: usize) -> Result<()> {
        self.current.advance(amount);
        Ok(())
    }

    fn read_parser<'a, O: 'a, P>(&'a mut self, mut parser: P) -> Result<O>
    where
        P: nom::Parser<&'a [u8], O, parse::Error<&'a [u8]>>,
    {
        // Store the current buffer to last so we can return a reference
        // to it.
        self.last = Some(self.current.clone());
        let buf = self.last.as_ref().unwrap();
        match parser.parse(buf.as_ref()) {
            Ok((rest, output)) => {
                // Modify current to refer to rest.
                // This is why parse last, as otherwise current would be borrowed here.
                self.current = buf.slice_ref(rest);
                Ok(output)
            }
            Err(nom::Err::Incomplete(_)) => bail!(UnexpectedEof),
            Err(nom::Err::Error(e) | nom::Err::Failure(e)) => Err(e.into_owned().into()),
        }
    }

    fn read_data_bytes(&mut self) -> Result<Bytes> {
        let len = self.read_data_length()?;
        Ok(self.current.split_to(len as usize))
    }
}
