use arrow2::{
    array::{
        MutableArray, MutableBinaryArray, MutableBooleanArray, MutableFixedSizeBinaryArray,
        MutableListArray, MutablePrimitiveArray, MutableStructArray, MutableUtf8ValuesArray,
        TryPush,
    },
    datatypes::{DataType, Field, TimeUnit},
    types::NativeType,
};
use rsbag::{
    layout::{FieldLayout, FieldType, MessageLayout, Multiplicity},
    parse::{header::fields::parse_time, Error, ErrorKind, IResult, Input},
    Result,
};
use eyre::bail;
use nom::{
    bytes::complete::take,
    combinator::map_res,
    multi::{length_count, length_data, many_m_n},
    number::complete::{
        le_f32, le_f64, le_i16, le_i32, le_i64, le_i8, le_u16, le_u32, le_u64, le_u8,
    },
    Parser,
};

fn make_mutable_array<A: MutableArray + Default + 'static>(
    multiplicity: Multiplicity,
) -> Box<dyn MutableArray> {
    make_mutable_array_from_values(multiplicity, A::default())
}

fn make_mutable_array_from_values<A: MutableArray + 'static>(
    multiplicity: Multiplicity,
    values: A,
) -> Box<dyn MutableArray> {
    match multiplicity {
        Multiplicity::Unit => Box::new(values),
        // TODO: cannot write fixed size list to parqeut
        // Multiplicity::Fixed(size) => Box::new(MutableFixedSizeListArray::new_with_field(
        //     values, "item", false, size,
        // )),
        Multiplicity::Fixed(_) | Multiplicity::Dynamic => Box::new(
            MutableListArray::<i32, A>::new_with_field(values, "item", false),
        ),
    }
}

fn field_to_mutable_array(field: &FieldLayout) -> Box<dyn MutableArray> {
    // TODO: BinaryArray for I8/U8
    match &field.typ {
        FieldType::Bool => make_mutable_array::<MutableBooleanArray>(field.multiplicity),
        FieldType::I8(_) => make_mutable_array::<MutablePrimitiveArray<i8>>(field.multiplicity),
        FieldType::I16 => make_mutable_array::<MutablePrimitiveArray<i16>>(field.multiplicity),
        FieldType::I32 => make_mutable_array::<MutablePrimitiveArray<i32>>(field.multiplicity),
        FieldType::I64 => make_mutable_array::<MutablePrimitiveArray<i64>>(field.multiplicity),
        FieldType::U8(_) => match field.multiplicity {
            Multiplicity::Unit => Box::new(MutablePrimitiveArray::<u8>::new()),
            Multiplicity::Fixed(n) => Box::new(MutableFixedSizeBinaryArray::new(n)),
            Multiplicity::Dynamic => Box::new(MutableBinaryArray::<i32>::new()),
        },
        FieldType::U16 => make_mutable_array::<MutablePrimitiveArray<u16>>(field.multiplicity),
        FieldType::U32 => make_mutable_array::<MutablePrimitiveArray<u32>>(field.multiplicity),
        FieldType::U64 => make_mutable_array::<MutablePrimitiveArray<u64>>(field.multiplicity),
        FieldType::F32 => make_mutable_array::<MutablePrimitiveArray<f32>>(field.multiplicity),
        FieldType::F64 => make_mutable_array::<MutablePrimitiveArray<f64>>(field.multiplicity),
        FieldType::String => make_mutable_array::<MutableUtf8ValuesArray<i32>>(field.multiplicity),
        FieldType::Time => make_mutable_array_from_values(
            field.multiplicity,
            MutablePrimitiveArray::<i64>::from(DataType::Timestamp(
                TimeUnit::Nanosecond,
                Some("UTC".into()),
            )),
        ),
        FieldType::Duration => make_mutable_array_from_values(
            field.multiplicity,
            MutablePrimitiveArray::<i64>::from(DataType::Duration(TimeUnit::Nanosecond)),
        ),
        FieldType::Message(layout) => {
            make_mutable_array_from_values(field.multiplicity, layout_to_mutable_array(&*layout))
        }
    }
}

fn layout_to_mutable_arrays(layout: &MessageLayout) -> Vec<Box<dyn MutableArray>> {
    layout.fields().iter().map(field_to_mutable_array).collect()
}

pub fn layout_to_mutable_array(layout: &MessageLayout) -> MutableStructArray {
    let values = layout_to_mutable_arrays(layout);
    let fields = layout
        .fields()
        .iter()
        .zip(&values)
        .map(|(field, array)| Field::new(field.name.to_string(), array.data_type().clone(), false))
        .collect();
    let datatype = DataType::Struct(fields);
    MutableStructArray::new(datatype, values)
}

// TODO: more efficient primitive parsing using bytemuck
fn parse_push_primitive<'a, 'b: 'a, T, P>(
    mut parser: P,
    array: &'a mut dyn MutableArray,
    multiplicity: Multiplicity,
    input: Input<'b>,
) -> IResult<'b, ()>
where
    T: NativeType,
    P: Parser<Input<'b>, T, Error<Input<'b>>> + 'a,
{
    match multiplicity {
        Multiplicity::Unit => {
            let array = array
                .as_mut_any()
                .downcast_mut::<MutablePrimitiveArray<T>>()
                .expect("wrong array type");
            let (rest, value) = parser.parse(input)?;
            array.push(Some(value));
            Ok((rest, ()))
        }
        Multiplicity::Fixed(n) => {
            // TODO: FixedSizeListArray unsupported to write to parquet
            // let array = array
            //     .as_mut_any()
            //     .downcast_mut::<MutableFixedSizeListArray<MutablePrimitiveArray<T>>>()
            //     .expect("wrong array type");
            let array = array
                .as_mut_any()
                .downcast_mut::<MutableListArray<i32, MutablePrimitiveArray<T>>>()
                .expect("wrong array type");
            let (rest, values) = many_m_n(n, n, parser).parse(input)?;
            array.mut_values().extend_from_slice(values.as_ref());
            array.try_push_valid().unwrap();
            Ok((rest, ()))
        }
        Multiplicity::Dynamic => {
            let array = array
                .as_mut_any()
                .downcast_mut::<MutableListArray<i32, MutablePrimitiveArray<T>>>()
                .expect("wrong array type");
            let (rest, values) = length_count(le_u32, parser).parse(input)?;
            array.mut_values().extend_from_slice(values.as_ref());
            array.try_push_valid().unwrap();
            Ok((rest, ()))
        }
    }
}

// TODO: more efficient primitive parsing using bytemuck
fn parse_push_u8<'a, 'b: 'a>(
    array: &'a mut dyn MutableArray,
    multiplicity: Multiplicity,
    input: Input<'b>,
) -> IResult<'b, ()> {
    // TODO: copy directly without
    match multiplicity {
        Multiplicity::Unit => {
            let array = array
                .as_mut_any()
                .downcast_mut::<MutablePrimitiveArray<u8>>()
                .expect("wrong array type");
            let (rest, value) = le_u8.parse(input)?;
            array.push(Some(value));
            Ok((rest, ()))
        }
        Multiplicity::Fixed(n) => {
            let array = array
                .as_mut_any()
                .downcast_mut::<MutableFixedSizeBinaryArray>()
                .expect("wrong array type");
            let (rest, values) = take(n).parse(input)?;
            array.push(Some(values));
            Ok((rest, ()))
        }
        Multiplicity::Dynamic => {
            let array = array
                .as_mut_any()
                .downcast_mut::<MutableBinaryArray<i32>>()
                .expect("wrong array type");
            let (rest, values) = length_data(le_u32).parse(input)?;
            array.push(Some(values));
            Ok((rest, ()))
        }
    }
}

fn parse_push<'a, 'b: 'a, A, T, P>(
    mut parser: P,
    array: &'a mut dyn MutableArray,
    multiplicity: Multiplicity,
    input: Input<'b>,
) -> IResult<'b, ()>
where
    P: Parser<Input<'b>, T, Error<Input<'b>>> + 'a,
    A: MutableArray + TryPush<T> + 'static,
{
    match multiplicity {
        Multiplicity::Unit => {
            let array = array
                .as_mut_any()
                .downcast_mut::<A>()
                .expect("wrong array type");
            let (rest, value) = parser.parse(input)?;
            array.try_push(value).expect("array push failed");
            Ok((rest, ()))
        }
        Multiplicity::Fixed(n) => {
            // TODO: FixedSizeListArray unsupported to write to parquet
            // let array = array
            //     .as_mut_any()
            //     .downcast_mut::<MutableFixedSizeListArray<A>>()
            //     .expect("wrong array type");
            let array = array
                .as_mut_any()
                .downcast_mut::<MutableListArray<i32, A>>()
                .expect("wrong array type");
            let (rest, values) = many_m_n(n, n, parser).parse(input)?;
            for value in values {
                array
                    .mut_values()
                    .try_push(value)
                    .expect("array push failed");
            }
            array.try_push_valid().unwrap();
            Ok((rest, ()))
        }
        Multiplicity::Dynamic => {
            let array = array
                .as_mut_any()
                .downcast_mut::<MutableListArray<i32, A>>()
                .expect("wrong array type");
            let (rest, values) = length_count(le_u32, parser).parse(input)?;
            for value in values {
                array
                    .mut_values()
                    .try_push(value)
                    .expect("array push failed");
            }
            array.try_push_valid().unwrap();
            Ok((rest, ()))
        }
    }
}

fn parse_push_message<'a, 'b: 'a>(
    layout: &MessageLayout,
    array: &'a mut dyn MutableArray,
    multiplicity: Multiplicity,
    input: Input<'b>,
) -> IResult<'b, ()> {
    match multiplicity {
        Multiplicity::Unit => {
            let array = array
                .as_mut_any()
                .downcast_mut::<MutableStructArray>()
                .expect("wrong array type");
            message_parser(layout, array).parse(input)
        }
        Multiplicity::Fixed(n) => {
            // TODO: FixedSizeListArray unsupported to write to parquet
            // let array = array
            //     .as_mut_any()
            //     .downcast_mut::<MutableFixedSizeListArray<MutableStructArray>>()
            //     .expect("wrong array type");
            let array = array
                .as_mut_any()
                .downcast_mut::<MutableListArray<i32, MutableStructArray>>()
                .expect("wrong array type");
            let parser = message_parser(layout, array.mut_values());
            let (rest, _) = many_m_n(n, n, parser).map(|_| ()).parse(input)?;
            array.try_push_valid().unwrap();
            Ok((rest, ()))
        }
        Multiplicity::Dynamic => {
            let array = array
                .as_mut_any()
                .downcast_mut::<MutableListArray<i32, MutableStructArray>>()
                .expect("wrong array type");
            let parser = message_parser(layout, array.mut_values());
            let (rest, _) = length_count(le_u32, parser).map(|_| ()).parse(input)?;
            array.try_push_valid().unwrap();
            Ok((rest, ()))
        }
    }
}

fn field_parser<'a, 'b: 'a>(
    field: &'a FieldLayout,
    array: &'a mut dyn MutableArray,
) -> impl Parser<Input<'b>, (), Error<Input<'b>>> + 'a {
    move |input| -> IResult<()> {
        match &field.typ {
            FieldType::Bool => parse_push::<MutableBooleanArray, _, _>(
                le_u8.map(|v| Some(v != 0)),
                array,
                field.multiplicity,
                input,
            ),
            FieldType::I8(_) => parse_push_primitive(le_i8, array, field.multiplicity, input),
            FieldType::I16 => parse_push_primitive(le_i16, array, field.multiplicity, input),
            FieldType::I32 => parse_push_primitive(le_i32, array, field.multiplicity, input),
            FieldType::I64 => parse_push_primitive(le_i64, array, field.multiplicity, input),
            FieldType::U8(_) => parse_push_u8(array, field.multiplicity, input),
            FieldType::U16 => parse_push_primitive(le_u16, array, field.multiplicity, input),
            FieldType::U32 => parse_push_primitive(le_u32, array, field.multiplicity, input),
            FieldType::U64 => parse_push_primitive(le_u64, array, field.multiplicity, input),
            FieldType::F32 => parse_push_primitive(le_f32, array, field.multiplicity, input),
            FieldType::F64 => parse_push_primitive(le_f64, array, field.multiplicity, input),
            FieldType::String => parse_push::<MutableUtf8ValuesArray<i32>, _, _>(
                map_res(length_data(le_u32), |s| {
                    String::from_utf8(Vec::from(s)).map_err(ErrorKind::from)
                }),
                array,
                field.multiplicity,
                input,
            ),
            FieldType::Time => parse_push_primitive(
                parse_time.map(|time| time.nanos()),
                array,
                field.multiplicity,
                input,
            ),
            FieldType::Duration => todo!(),
            FieldType::Message(layout) => {
                parse_push_message(layout.as_ref(), array, field.multiplicity, input)
            }
        }
    }
}

fn message_parser<'a, 'b: 'a>(
    layout: &'a MessageLayout,
    array: &'a mut MutableStructArray,
) -> impl Parser<Input<'b>, (), Error<Input<'b>>> + 'a {
    move |mut input| -> IResult<_> {
        for (field, values) in layout.fields().iter().zip(array.mut_values()) {
            let (rest, _) = field_parser(field, &mut **values).parse(input)?;
            input = rest;
        }
        Ok((input, ()))
    }
}

pub fn parse_message(
    layout: &MessageLayout,
    input: &[u8],
    array: &mut MutableStructArray,
) -> Result<()> {
    match message_parser(layout, array).parse(input) {
        Ok((&[], message)) => Ok(message),
        Ok(_) => bail!("extra data after message"),
        Err(nom::Err::Incomplete(_)) => unreachable!(),
        Err(nom::Err::Error(e) | nom::Err::Failure(e)) => Err(e.into_owned().into()),
    }
}
