use std::{
    collections::{BTreeMap, BTreeSet, HashMap, VecDeque},
    env::args,
    fs::File,
    path::Path,
    sync::{Arc, Mutex},
};

use arrow2::{
    array::{Array, MutableArray, MutableStructArray},
    chunk::Chunk,
    datatypes::{DataType, Field, PhysicalType, Schema},
    io::parquet::write::{self as pq_write, row_group_iter, CompressedPage, ZstdLevel},
    types::PrimitiveType,
};
use indicatif::ParallelProgressIterator;
use log::{info, trace};
use rayon::{
    iter::ParallelIterator,
    prelude::{IndexedParallelIterator, ParallelBridge},
};
use rsbag::{
    chunk::{read_chunk_data_at, ChunkMessageIterator, MessageData},
    index::BagIndex,
    layout::MessageLayout,
    message::MessageLayouts,
    Bag, Result,
};

#[cfg(not(target_env = "msvc"))]
use jemallocator::Jemalloc;

#[cfg(not(target_env = "msvc"))]
#[global_allocator]
static GLOBAL: Jemalloc = Jemalloc;

mod arrow;

struct Bla {
    columns: VecDeque<CompressedPage>,
    current: Option<CompressedPage>,
}

impl Bla {
    pub fn new(columns: VecDeque<CompressedPage>) -> Self {
        Self {
            columns,
            current: None,
        }
    }
}

impl pq_write::FallibleStreamingIterator for Bla {
    type Item = CompressedPage;
    type Error = arrow2::error::Error;

    fn advance(&mut self) -> Result<(), arrow2::error::Error> {
        self.current = self.columns.pop_front();
        Ok(())
    }

    fn get(&self) -> Option<&Self::Item> {
        self.current.as_ref()
    }
}

type ChunkSet = BTreeSet<usize>;

struct ConnToArrow {
    layout: MessageLayout,
    array: MutableStructArray,
    expected_chunks: ChunkSet,
    /// Messages waiting to be inserted in array in order of chunk
    pending_messages: BTreeMap<usize, Vec<MessageData>>,
    estimated_size: usize,
    ready_chunks: VecDeque<Chunk<Box<dyn Array>>>,
    max_chunk_size: usize,
}

impl ConnToArrow {
    fn try_new(layout: MessageLayout, expected_chunks: ChunkSet) -> Result<Self> {
        let array = arrow::layout_to_mutable_array(&layout);

        Ok(Self {
            layout: layout,
            array: array,
            expected_chunks,
            pending_messages: Default::default(),
            estimated_size: 0,
            ready_chunks: Default::default(),
            max_chunk_size: 128 * 1024 * 1024,
        })
    }

    fn add_message(&mut self, chunk_idx: usize, message: MessageData) {
        self.pending_messages
            .entry(chunk_idx)
            .or_default()
            .push(message);
    }

    fn chunk_complete(&mut self, chunk_idx: usize) -> Result<()> {
        assert!(self.expected_chunks.remove(&chunk_idx));
        // Flush all pending messages from chunks before the last expected chunk.
        // i.e. wait to flush a chunk until all previous chunks have been recieved.
        let latest_expected_chunk = self.expected_chunks.first().copied();
        let mut flushed = 0;
        for (&chunk_idx, messages) in self.pending_messages.iter_mut() {
            if let Some(latest_expected_chunk) = latest_expected_chunk {
                if chunk_idx >= latest_expected_chunk {
                    break;
                }
            }
            flushed += 1;
            for message in messages.drain(..) {
                arrow::parse_message(&self.layout, message.data.as_ref(), &mut self.array)?;
                self.estimated_size += message.data.len();
            }
        }
        trace!(
            "Chunk {} complete, with {} pending chunks (latest {:?}). Flushed {} chunks",
            chunk_idx,
            self.pending_messages.len(),
            latest_expected_chunk,
            flushed
        );
        // Remove all empty chunks
        self.pending_messages.retain(|_, v| !v.is_empty());

        if self.estimated_size > self.max_chunk_size {
            trace!(
                "connection data size exceeds {}MiB, flushing",
                self.max_chunk_size / (1024 * 1024)
            );
            self.flush();
        }

        Ok(())
    }

    fn flush(&mut self) {
        let chunk = Chunk::new(
            std::mem::take(self.array.mut_values())
                .into_iter()
                .map(|mut v| v.as_box())
                .collect(),
        );
        self.ready_chunks.push_back(chunk);

        self.array = arrow::layout_to_mutable_array(&self.layout);
        self.estimated_size = 0;
    }

    fn take_chunk(&mut self) -> Option<Chunk<Box<dyn Array>>> {
        self.ready_chunks.pop_front()
    }
}

struct ParquetWriter {
    encodings: Vec<Vec<pq_write::Encoding>>,
    file_writer: pq_write::FileWriter<File>,
}

impl ParquetWriter {
    fn try_new(fields: &[Field], file: File) -> Result<Self> {
        let writer_options = pq_write::WriteOptions {
            write_statistics: true,
            version: pq_write::Version::V2,
            compression: pq_write::CompressionOptions::Zstd(Some(ZstdLevel::try_new(10).unwrap())),
            // compression: pq_write::CompressionOptions::Zstd(None),
            // compression: pq_write::CompressionOptions::Snappy,
            // compression: pq_write::CompressionOptions::Uncompressed,
            data_pagesize_limit: Some(128 * 1024 * 1024),
            // data_pagesize_limit: Some(16 * 1024 * 1024),
        };
        let schema = Schema {
            fields: fields.to_vec(),
            metadata: Default::default(),
        };

        let encodings: Vec<_> = fields
            .iter()
            .map(|f| {
                pq_write::transverse(&f.data_type, |dtype| match dtype.to_physical_type() {
                    PhysicalType::Primitive(pt) => match pt {
                        PrimitiveType::Float16
                        | PrimitiveType::Float32
                        | PrimitiveType::Float64 => pq_write::Encoding::Plain,
                        _ => pq_write::Encoding::DeltaBinaryPacked,
                    },
                    PhysicalType::Binary => pq_write::Encoding::DeltaLengthByteArray,
                    _ => pq_write::Encoding::Plain,
                })
            })
            .collect();

        let file_writer = pq_write::FileWriter::try_new(file, schema, writer_options)?;
        Ok(Self {
            encodings,
            file_writer,
        })
    }

    fn write_chunk(&mut self, chunk: Chunk<Box<dyn Array>>) -> Result<()> {
        let options = self.file_writer.options();
        // let columns = chunk
        //     .columns()
        //     .par_iter()
        //     .zip(self.file_writer.parquet_schema().fields().to_vec())
        //     .zip(self.encodings.par_iter())
        //     .flat_map(move |((array, type_), encoding)| {
        //         let encoded_columns =
        //             pq_write::array_to_columns(array, type_, options, encoding).unwrap();
        //         encoded_columns
        //             .into_iter()
        //             .map(|encoded_pages| {
        //                 let encoded_pages = pq_write::DynIter::new(
        //                     encoded_pages
        //                         .into_iter()
        //                         .map(|x| x.map_err(arrow2::error::Error::from_external_error)),
        //                 );
        //                 encoded_pages
        //                     .map(|page| {
        //                         pq_write::compress(page?, vec![], options.compression)
        //                             .map_err(|x| x.into())
        //                     })
        //                     .collect::<Result<VecDeque<_>>>()
        //             })
        //             .collect::<Vec<_>>()
        //     })
        //     .collect::<Result<Vec<VecDeque<pq_write::CompressedPage>>>>()?;

        // let rgi = pq_write::DynIter::new(
        //     columns
        //         .into_iter()
        //         .map(|column| Ok(pq_write::DynStreamingIterator::new(Bla::new(column)))),
        // );

        let rgi = row_group_iter(
            chunk,
            self.encodings.clone(),
            self.file_writer.parquet_schema().fields().to_vec(),
            options,
        );
        self.file_writer.write(rgi)?;

        Ok(())
    }

    fn end(&mut self) -> Result<()> {
        // TODO: footer metadata?
        self.file_writer.end(None)?;
        Ok(())
    }
}

struct ConnToParquet {
    to_arrow: Mutex<ConnToArrow>,
    parquet_writer: Mutex<ParquetWriter>,
}

impl ConnToParquet {
    fn try_new(layout: MessageLayout, expected_chunks: ChunkSet, file: File) -> Result<Self> {
        let to_arrow = ConnToArrow::try_new(layout, expected_chunks)?;

        let DataType::Struct(fields) = to_arrow.array.data_type() else { unreachable!() };
        let parquet_writer = ParquetWriter::try_new(fields, file)?;

        Ok(Self {
            to_arrow: to_arrow.into(),
            parquet_writer: parquet_writer.into(),
        })
    }

    fn add_message(&self, chunk_idx: usize, message: MessageData) {
        self.to_arrow
            .lock()
            .unwrap()
            .add_message(chunk_idx, message);
    }

    fn chunk_complete(&self, chunk_idx: usize) -> Result<()> {
        self.to_arrow.lock().unwrap().chunk_complete(chunk_idx)?;

        self.write_ready()?;

        Ok(())
    }

    fn write_ready(&self) -> Result<()> {
        loop {
            let Some(chunk) = ({
                let mut to_arrow = self.to_arrow.lock().unwrap();
                to_arrow.take_chunk()
            }) else { break };

            self.parquet_writer.lock().unwrap().write_chunk(chunk)?;
        }
        Ok(())
    }

    fn end(&self) -> Result<()> {
        self.to_arrow.lock().unwrap().flush();
        self.write_ready()?;
        self.parquet_writer.lock().unwrap().end()?;
        Ok(())
    }
}

struct BagToParquet {
    connections: HashMap<u32, ConnToParquet>,
    index: BagIndex,
}

impl BagToParquet {
    fn try_new(index: &BagIndex, layouts: &MessageLayouts, base_path: &Path) -> Result<Self> {
        let mut expected_chunks_per_conn: HashMap<u32, ChunkSet> = Default::default();
        for (chunk_idx, chunk) in index.chunks.iter().enumerate() {
            for conn in &chunk.connections {
                let expected_chunks = expected_chunks_per_conn.entry(conn.conn_id).or_default();
                expected_chunks.insert(chunk_idx);
            }
        }
        let connections: HashMap<_, _> = layouts
            .conns
            .iter()
            .filter(|(_, conn)| {
                !conn.info.topic.ends_with("image_rect_raw")
                    || conn.info.topic.starts_with("/camera3")
            })
            .map(|(id, conn)| -> Result<_> {
                let expected_chunks = std::mem::take(expected_chunks_per_conn.get_mut(id).unwrap());
                let conn_path = base_path.join(format!("{}.parquet", &conn.info.topic[1..]));
                if let Some(parent) = conn_path.parent() {
                    mkdirp::mkdirp(parent)?;
                }
                let conn_file = File::create(conn_path)?;
                let ctp = ConnToParquet::try_new(conn.layout.clone(), expected_chunks, conn_file)?;
                Ok((*id, ctp))
            })
            .collect::<Result<_>>()?;
        Ok(Self {
            connections,
            index: index.clone(),
        })
    }

    fn add_message(&self, chunk_idx: usize, message: MessageData) {
        if let Some(ctp) = self.connections.get(&message.header.conn_id) {
            ctp.add_message(chunk_idx, message)
        }
    }

    fn chunk_complete(&self, chunk_idx: usize) -> Result<()> {
        let chunk_info = self.index.chunks.get(chunk_idx).unwrap();
        for conn in &chunk_info.connections {
            if let Some(ctp) = self.connections.get(&conn.conn_id) {
                ctp.chunk_complete(chunk_idx)?;
            }
        }
        Ok(())
    }

    fn end(&self) -> Result<()> {
        self.connections
            .values()
            .par_bridge()
            .try_for_each(|ctp| ctp.end())?;
        Ok(())
    }
}

fn main() -> Result<()> {
    color_eyre::install()?;
    env_logger::init();

    let args: Vec<_> = args().collect();
    if args.len() != 3 {
        eprintln!("Usage: {} <bag path> <parquet file>", args[0]);
        return Ok(());
    }

    let bag_path = &args[1];
    let arrow_path = &args[2];
    let mut bag = Bag::open(bag_path)?;

    let layouts = bag.compute_message_layouts()?;

    // let info = bag.compute_info()?;
    // let total_messages: u64 = info.per_connection.values().map(|con| con.count).sum();
    // info!("exporting {} messages", total_messages);

    let num_chunks = bag.index().chunks.len();
    let btp = Arc::new(BagToParquet::try_new(
        bag.index(),
        &layouts,
        Path::new(arrow_path),
    )?);
    bag.index()
        .chunks
        .iter()
        .enumerate()
        // Force using par_bridge as into_par_iter distributes chunks throughout the file
        .par_bridge()
        .progress_count(num_chunks as u64)
        .map_with(bag.reader().clone(), |reader, (idx, chunk)| {
            (idx, read_chunk_data_at(reader, chunk.pos))
        })
        .for_each_with(btp.clone(), |btp, (chunk_idx, chunk)| {
            let it = ChunkMessageIterator::new(chunk.unwrap());
            for msg in it {
                let msg = msg.unwrap();
                btp.add_message(chunk_idx, msg);
            }
            btp.chunk_complete(chunk_idx).unwrap();
        });

    btp.end()?;

    Ok(())
}
